/* This file is Copyright 2001 Level Control Systems.  See the included LICENSE.TXT file for details. */
package com.lcs.muscle.client;

import java.net.Socket;
import java.io.DataOutput;
import java.io.DataInput;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.IOException;

import com.lcs.muscle.iogateway.AbstractMessageIOGateway;
import com.lcs.muscle.iogateway.MessageIOGateway;
import com.lcs.muscle.message.Message;
import com.lcs.muscle.support.LEDataInputStream;
import com.lcs.muscle.support.LEDataOutputStream;
import com.lcs.muscle.thread.MessageListener;
import com.lcs.muscle.thread.MessageQueue;

/** This class is a handy utility class that acts as
 *  an interface to a TCP socket to send and receive
 *  Messages over.
 */
public class MessageTransceiver implements MessageListener 
{
   /** Start up a transceiver that will connect to the given ipAddress and port,
    *  and send status messages and Messages to (repliesTo).
    *  @param repliesTo where to send status & reply messages
    */
   public MessageTransceiver(MessageQueue repliesTo) 
   {      
      _repliesTo = repliesTo;
      _gateway   = new MessageIOGateway();  // default protocol
      _sendQueue = new MessageQueue(new MessageTransmitter());      
   }

   /** Start up a transceiver that will connect to the given ipAddress and port,
    *  and send status messages and Messages to (repliesTo).
    *  @param repliesTo where to send status & reply messages
    *  @param ioGateway the AbstractMessageIOGateway to use to flatten and unflatten Messages.
    */
   public MessageTransceiver(MessageQueue repliesTo, AbstractMessageIOGateway ioGateway)
   {
      _repliesTo = repliesTo;
      _gateway   = ioGateway;
      _sendQueue = new MessageQueue(new MessageTransmitter());
   }
   
   /** Tells the internal thread to connect to the given host and port
    *  Connection is done asynchronously, and an OperationCompleteMessage
    *  will be sent to the repliesTo queue when it has been done.
    *  @param hostName IP address or host name of computer to connect to
    *  @param port Port number to connect to (2960 is the port for a MUSCLE server)
    *  @param successTag Any value you want--if non-null, this will be sent back to indicate that the connection succeeded.
    *  @param failTag Any value you want--if non-null, this will be sent back to indicate that the connection failed or was broken.
    */
   public void connect(String hostName, int port, Object successTag, Object failTag)
   {
      _sendQueue.postMessage(new ConnectMessage(hostName, port, successTag, failTag));
   }
   
   /** Tells the internal thread to disconnect.  This will be done asynchronously. 
    *  When the disconnection is done, the failTag specified previously in the connect()
    *  call will be sent to your reply queue.
    */
   public void disconnect()
   {
      _sendQueue.postMessage(new ConnectMessage(null, 0, null, null));
   }
    
   /** Tells the internal thread to send a Message out over the TCP socket.
    *  This will be done asynchronously.
    */
   public void sendOutgoingMessage(Message msg)
   {
      _sendQueue.postMessage(msg);
   }
   
   /** Any received Messages will be sent out to the TCP socket */
   public void messageReceived(Object msg, int numLeft)
   {
      if (msg instanceof Message) sendOutgoingMessage((Message)msg);
   }

   /** Internal control class */
   private final class ConnectMessage    
   {
      public ConnectMessage(String hostName, int port, Object successTag, Object failTag)
      {
         _hostName   = hostName;
         _port       = port;
         _successTag = successTag;
         _failTag    = failTag;
      }
      
      public String _hostName;
      public int    _port;
      public Object _successTag;
      public Object _failTag;
   }
   
   /** Logic for the connecting-and-sending thread */
   private final class MessageTransmitter implements MessageListener
   {
      public MessageTransmitter()
      {
         // empty
      }
 
      public void messageReceived(Object message, int numLeft)
      {
         if (message instanceof Message)
         {
            if (_out != null) 
            {
               try {
                  _gateway.flattenMessage(_out, (Message) message);
               }
               catch(IOException ex) {
                  disconnect(_connectionID);
               }       
            }
         }
         else if (message instanceof ConnectMessage)
         {
            disconnect(_connectionID);  // make sure our previous connection is gone
            
            ConnectMessage cmsg = (ConnectMessage) message;
            if (cmsg._hostName != null)
            {
               // Do the actual connecting in Yet Another Thread, so that we can remain
               // responsive to the user's requests during the (possibly lengthy) connect period.
               _currentConnectThread = new ConnectThread(cmsg);  // connect thread is started in ctor
            }
         }
         else if (message instanceof ConnectThread)
         {
            // When our ConnectThread has finished, he sends himself back to us with the connect results.
            ConnectThread ct = (ConnectThread) message;
            Socket s = ct._socket;
            ConnectMessage cmsg = ct._cmsg;
            if (ct == _currentConnectThread)
            {
               _currentConnectThread = null;   // no longer in the connect period
               if (s != null)
               {
                  try {
                     _socket  = s;
                     _failTag = cmsg._failTag;
                     _out     = new LEDataOutputStream(new DataOutputStream(new BufferedOutputStream(_socket.getOutputStream())));
                     new MessageReceiver(new LEDataInputStream(new DataInputStream(new BufferedInputStream(_socket.getInputStream()))), _connectionID);  // start the read thread
                     sendReply(cmsg._successTag);  // ok, done!
                  }
                  catch(IOException ex) {
                     disconnect(_connectionID);
                  }
               }
               else sendReply(cmsg._failTag);               
            }
            else 
            {
               // oops, no use for the socket, so just close it and forget about it
               if (s != null) try {s.close();} catch(Exception ex) {/* don't care */}
               sendReply(cmsg._failTag);
            }
         }
         else if (message instanceof MessageReceiver) disconnect(((MessageReceiver)message).connectionID());

         if ((numLeft == 0)&&(_out != null)) {
            try {
               _out.flush();
            }
            catch(IOException ex) {
               ex.printStackTrace(); 
            }
         }
      }

      /** Close the connection and notify the user (only if the connection ID is correct) */
      private void disconnect(int id)
      {
         if ((_socket != null)&&(_connectionID == id))
         {
            try {
               _socket.close();
            }
            catch(IOException ex) {
               // ignore
            }
            sendReply(_failTag);
            _out = null;
            _socket = null;
            _failTag = null;
            _currentConnectThread = null;
            _connectionID++;
         }
      }
      
      /** Send a message back to the user */
      private void sendReply(Object replyObj)
      {
         if ((_repliesTo != null)&&(replyObj != null)) _repliesTo.postMessage(replyObj);
      }

      /** Logic for our receive-incoming-data thread */
      private class MessageReceiver implements MessageListener
      {
         public MessageReceiver(DataInput in, int cid)
         {
            _in = in;
            _connectionID = cid;
            (new MessageQueue(this)).postMessage(null);  // start the background reader thread
         }
         
         public void messageReceived(Object msg, int numLeft)
         {
            try {
               while(true) sendReply(_gateway.unflattenMessage(_in));
            }
            catch(IOException ex) {
               // oops, socket broke... notify out master that we're leaving (by posting ourself as a message, teehee)
               _sendQueue.postMessage(this);
            }              
         }
         
         public int connectionID() {return _connectionID;}
         
         private int _connectionID;
         private DataInput _in;
      }
      

      private Object _failTag         = null;
      private Socket _socket          = null;
      private LEDataOutputStream _out = null;
   }

   // A little class for doing TCP connects in a separate thread (so the MessageTransceiver won't block during the connect) */
   private class ConnectThread implements MessageListener
   {
      public ConnectThread(ConnectMessage cmsg)
      {
         _cmsg = cmsg;
         (new MessageQueue(this)).postMessage(null);  // just to kick off the thread
      }

      public void messageReceived(Object obj, int numLeft)
      {
         try {
            _socket = new Socket(_cmsg._hostName, _cmsg._port);         
         }
         catch(IOException ex) {
            // do nothing
         }
         _sendQueue.postMessage(this);
      }
      
      public Socket _socket = null;
      public ConnectMessage _cmsg;
   }

   private AbstractMessageIOGateway _gateway;
   private MessageQueue _repliesTo;   
   private MessageQueue _sendQueue;
   private int _connectionID = 0;
   private ConnectThread _currentConnectThread = null;  // non-null iff we are connecting (in a slave thread)
}
