"""
This class represents an asynchronous TCP connection
over which MUSCLE Messages may be sent and received
without blocking the main thread of execution.
"""

__author__    = "Jeremy Friesner (jaf@lcsaudio.com)"
__version__   = "$Revision: 1.6 $"
__date__      = "$Date: 2002/04/26 06:34:04 $"
__copyright__ = "Copyright (c) 2002 Level Control Systems"
__license__   = "See enclosed LICENSE.TXT file"

import Message
import threading
import Queue
import select
import socket
import cStringIO
import struct
import sys

# constants used in our notifications
MTT_EVENT_CONNECTED     = 0
MTT_EVENT_DISCONNECTED  = 1

# Used internally by the MessageTransceiverThread class, to clean up
_EndSession = "endsession"

# Muscle gateway protocol's magic cookie, for sanity checking
MUSCLE_MESSAGE_ENCODING_DEFAULT = 1164862256  # 'Enc0'

# stolen from asyncore.py, and I'm not sorry
import os
if os.name == 'nt':
    EWOULDBLOCK = 10035
    EINPROGRESS = 10036
    EALREADY    = 10037
    ECONNRESET  = 10054
    ENOTCONN    = 10057
    ESHUTDOWN   = 10058
else:
    from errno import EALREADY, EINPROGRESS, EWOULDBLOCK, ECONNRESET, ENOTCONN, ESHUTDOWN 


class MessageTransceiverThread(threading.Thread):
   """Python implementation of the MUSCLE MessageTransceiverThread class.

   This class represents an asynchronous TCP connection over which MUSCLE Messages may be sent and received
   asynchronously, without ever blocking the main thread of execution.  When the TCP connection is established, 
   you will be notified with an MTT_EVENT_CONNECTED event.  When the TCP connection fails or is broken, you will
   be notified with an MTT_EVENT_DISCONNECTED event.  Whenever a Message is received over the TCP connection,
   you will get a Message object as your notification.  (See GetNextIncomingEvent() for details on notification).
   """

   def __init__(self, hostname, port, acceptFrom=None):
      """This constructor creates a MessageTransceiverThread that does an outgoing TCP connection.

      (hostname) is the hostname (or IP address string) of the host to connect to (e.g. "beshare.befaqs.com"
                 or "132.239.50.8").  If (hostname) is set to None, then this MessageTransceiverThread will 
                 be configured to accept incoming TCP connections, instead of initiating an outgoing one.
      (port) is the port number to connect to or accept on (e.g. 2960 is the default MUSCLE server port).
             If you are accepting, and (port) is set to zero, the system will choose a port number for you.
             (use GetPort() to find out which one it chose).
      (acceptfrom) is only relevant if you are accepting incoming connections, i.e. if (hostname) was None.
                   If specified, it is the only IP address that is allowed to connect to you.  If left as
                   None, however, any IP address is allowed to connect to you.

      Note that you will need to also call start() before the internal TCP connection thread will start executing."""

      threading.Thread.__init__(self)
      self.__outQ         = Queue.Queue(0)
      self.__inQ          = Queue.Queue(0)
      self.__hostname     = hostname
      self.__port         = port
      self.__acceptsocket = None
      self.__mainsocket   = None
      self.__threadsocket = None

      # Set up for accepting incoming TCP connections, if necessary
      if self.__hostname == None:
         if acceptFrom == None:
            acceptFrom = ""
         self.__acceptsocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
         self.__acceptsocket.bind((acceptFrom, port))
         self.__acceptsocket.listen(1)
         self.__acceptsocket.setblocking(0)
         self.__port = self.__acceptsocket.getsockname()[1]

      # Set up our internal socket-based thread notification/wakeup system
      acceptsock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
      acceptsock.bind(("127.0.0.1", 0))
      acceptsock.listen(1)
      self.__mainsocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
      self.__mainsocket.connect(acceptsock.getsockname())
      self.__threadsocket = acceptsock.accept()[0]
      acceptsock.close()
      self.__threadsocket.setblocking(0)
      self.__mainsocket.setblocking(0)

   def SendOutgoingMessage(self, msg):
      """Call this to queue up the given Message object to be sent out across the network.

      This method may be called at any time, but queued Messages won't go out until after
      the internal thread is start()'d and the TCP connection is established.  All Messages will
      be sent out in the same order that they were enqueued (FIFO).  This method is thread safe,
      and is typically called by your main thread of execution."""

      self.__outQ.put(msg, 0)
      self.__mainsocket.send("j")  # wake up the internal thread to check the Q

   def Destroy(self):
      """Shuts down the internal thread, closes any TCP connection and releases all held resources.

         Call this method when you are done using this object forever and want it to go away.  
         This method is typically called by your main thread of execution."""
      self.SendOutgoingMessage(_EndSession)  # special 'please die' code
      self.join()                            # wait for the internal thread to go away
      self.__outQ = None
      self.__inQ  = None
      if self.__mainsocket != None:
         self.__mainsocket.close()
      if self.__threadsocket != None:
         self.__threadsocket.close()
      if self.__acceptsocket != None:
         self.__aceeptsocket.close()

   def GetPort(self):
      """Returns the port this MessageTransceiverThread is connecting or accepting connections on."""
      return self.__port

   def GetNextIncomingEvent(self, block=0):
      """Returns the next incoming event object, or None if the incoming-event-queue is empty.  

         Incoming event objects will be one of the following three things:
            MTT_EVENT_CONNECTED    - TCP connection has been established
            MTT_EVENT_DISCONNECTED - TCP connection has been broken or failed
            (a Message object)     - A Message object that was received over the network.

         If (block) is set to true, this method will block until the next event
         is ready.  Otherwise (the default) this method will always return immediately.
         THIS METHOD IS THREAD SAFE, YOU MAY CALL IT AT ANY TIME, FROM ANY THREAD."""
      try:
         return self.__inQ.get(block)
      except Queue.Empty:
         return None

   def GetNotificationSocket(self):
      """Returns a socket that your main thread may select() or recv() on to wait for incoming network events.  

         Whenever a new event is ready, a byte will be sent to this socket.  Recv() it, then call 
         GetNextIncomingEvent() repeatedly to get all the new events, until it returns None.  
         Note:  please do not close() or write to this socket; it is still owned by (and used by) 
                this MessageTransceiverThread object."""
      return self.__mainsocket

   def NotifyMainThread(self):
      """This method is called by the MessageTransceiverThread's internal thread whenever 
         it wants to notify the main thread that there is a new event ready for the main
         thread to process.  This method's default behaviour is to send a byte to the 
         notification socket (see GetNotificationSocket(), above), which will wake up 
         the main thread if the main thread is blocking on that socket (via recv() or
         select() or whatnot).  However, if you wish to wake up the main thread using a
         different mechanism, feel free to override this method to do something more appropriate."""
      self.__threadsocket.send("t")

   def run(self):
      """Entry point for the internal thread.  Don't call this method; call start() instead."""
      toremote = None
      try:
         connectStillInProgress = 0
         if self.__acceptsocket == None:
            toremote = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            toremote.setblocking(0)
            try:
               toremote.connect((self.__hostname, self.__port))
            except socket.error, why:
               if why[0] in (EINPROGRESS, EALREADY, EWOULDBLOCK):
                  connectStillInProgress = 1
               else:
                  raise
            if connectStillInProgress == 0:
               self.__sendReplyToMainThread(MTT_EVENT_CONNECTED)

         # setup the lists we will select() on in advance, to avoid having to do it every time
         inlist       = [toremote, self.__threadsocket]
         notifyonly   = [self.__threadsocket]
         notifyaccept = [self.__threadsocket, self.__acceptsocket]
         outlist      = [toremote]
         emptylist    = []

         inHeader, inBody, inBodySize = "", "", 0
         outHeader, outBody = self.__getNextMessageFromMain()
         while 1:
            try:
               waitingForAccept = self.__acceptsocket != None and toremote == None

               if (outHeader != None or outBody != None or connectStillInProgress) and waitingForAccept == 0:
                  useoutlist = outlist
               else:
                  useoutlist = emptylist

               if connectStillInProgress:
                  useinlist = notifyonly
               elif waitingForAccept:
                  useinlist = notifyaccept
               else:
                  useinlist = inlist

               inready, outready, exlist = select.select(useinlist, useoutlist, emptylist)

               if waitingForAccept and self.__acceptsocket in inready:
                  toremote = self.__acceptsocket.accept()[0]
                  print "toremote is now ", toremote
                  inlist  = [toremote, self.__threadsocket]
                  outlist = [toremote]
                  self.__sendReplyToMainThread(MTT_EVENT_CONNECTED)

               if toremote in inready:
                  if inHeader != None:
                     newbytes = toremote.recv(8-len(inHeader))
                     if len(newbytes) == 0:
                        raise socket.error
                     inHeader = inHeader + newbytes
                     if len(inHeader) == 8:   # full header size is 2*sizeof(int32)
                        # Okay, we have the entire inHeader!  Now decode it
                        inBodySize, magic = struct.unpack("<2L", inHeader)
                        if magic != MUSCLE_MESSAGE_ENCODING_DEFAULT:
                           raise socket.error
                        inHeader = None    # signal that we are now ready to read the inBody
                  else:
                     newbytes = toremote.recv(inBodySize-len(inBody))
                     if len(newbytes) == 0:
                        raise socket.error
                     inBody = inBody + newbytes
                     if len(inBody) == inBodySize:
                        inmsg = Message.Message()
                        inmsg.Unflatten(cStringIO.StringIO(inBody))
                        self.__sendReplyToMainThread(inmsg)
                        inHeader, inBody, inBodySize = "", "", 0 # reset to receive next inHeader

               # Note that we don't need to do anything with the data we read from
               # threadsocket; it's enough that it woke us up so we can check the out-queue.
               if self.__threadsocket in inready and len(self.__threadsocket.recv(1024)) == 0:
                     raise _EndSession

               if toremote in outready:
                  if connectStillInProgress:
                     toremote.send("") # finalize the connect          # everyone else can get by with just this
                     self.__sendReplyToMainThread(MTT_EVENT_CONNECTED)
                     connectStillInProgress = 0
                  else:
                     if outHeader != None:
                        numSent = toremote.send(outHeader)
                        if numSent > 0:
                           outHeader = outHeader[numSent:]
                           if len(outHeader) == 0:
                              outHeader = None
                     elif outBody != None:
                        numSent = toremote.send(outBody)
                        if numSent > 0:
                           outBody = outBody[numSent:]
                           if len(outBody) == 0:
                              outBody = None

               if outHeader == None and outBody == None:
                  outHeader, outBody = self.__getNextMessageFromMain()
            except socket.error:
               if self.__acceptsocket != None:
                  self.__sendReplyToMainThread(MTT_EVENT_DISCONNECTED)
                  toremote = None # for accepting sessions, we can disconnect and then go back to waiting for an accept
               else:
                  raise # for connecting sessions, the first disconnection means game over, so rethrow
      except socket.error:
         self.__sendReplyToMainThread(MTT_EVENT_DISCONNECTED)
      except _EndSession:
         pass   # this exception just means to end the thread without comment

      if toremote != None:
         toremote.close()

   def __sendReplyToMainThread(self, what):
      self.__inQ.put(what, 0)
      self.NotifyMainThread()

   def __getNextMessageFromMain(self):
      try:
         next = self.__outQ.get(0)
         if next == _EndSession:
            raise _EndSession
         sio = cStringIO.StringIO()
         next.Flatten(sio)
         return (struct.pack("<2L", next.FlattenedSize(), MUSCLE_MESSAGE_ENCODING_DEFAULT), sio.getvalue())
      except Queue.Empty:
         return None, None  # oops, nothing more to send for the moment




# -----------------------------------------------------------------------------------------------
#
# Test stub below: exercises basic sending and receiving of Message objects.
# Easiest way to test this is to open two shell windows, and in the first type:
#   python MessageTransceiverThread.py accept 9999
# and then in the other shell window, type:
#   python MessageTransceiverThread.py localhost 9999
# Then the two programs should connect to each other, and typing anything into
# one window should result in output being printed to the other.  Fun!
#
# -----------------------------------------------------------------------------------------------

if __name__ == "__main__":

   # A custom subclass to make things simple for us.  A "real" program wouldn't do it this way.
   class TestMTT(MessageTransceiverThread):
      def __init__(self, hostname, port):
         MessageTransceiverThread.__init__(self, hostname, port)

      # Overridden to merely dequeue and print out any new incoming Messages or events.
      # Note that this isn't usually the best way to handle incoming events in a 'real' 
      # program, since this method is executed in the internal network thread and not in
      # the main thread.  In a more complex program, doing stuff here could easily lead
      # to race conditions.  Therefore, a better-written program would notify the main 
      # thread, and the main thread would respond by picking up the queued events and    
      # handle them locally, instead.  But this will do, for testing.
      # (see pythonchat.py for an example of a program that does things the right way)
      def NotifyMainThread(self):
         while 1:
            next = self.GetNextIncomingEvent()
            if next == MTT_EVENT_CONNECTED:
               print ""
               print "TCP Connection has been successfully created!"
            elif next == MTT_EVENT_DISCONNECTED:
               print ""
               print "TCP Connection failed or was disconnected!"
            elif next == None:
               break  # If we got here, the Queue is empty, nothing more to show
            else:
               print ""
               print "Received the following Message from the remote peer:"
               next.PrintToStream()
      

   if len(sys.argv) < 2:
      print "Usage:  python MessageTransceiverThread.py accept [port=0]"
      print "   or:  python MessageTransceiverThread.py <hostname> [port=2960]"
      sys.exit(5)

   if sys.argv[1] == "accept":
      port = 0   # Default -- 0 means we get to pick the port
      if len(sys.argv) >= 3:
         port = sys.argv[2]
      mtt = TestMTT(None, int(port))
      print "Accepting connections on port", mtt.GetPort()
   else:
      hostname = sys.argv[1]
      port     = 2960
      if len(sys.argv) >= 3:
         port = sys.argv[2]
      mtt = TestMTT(hostname, int(port))
      print "Attempting to connect to", hostname, "port", port, "..."

   mtt.start();  # important!  Otherwise nothing will happen :^)
   while 1:
      nextline = sys.stdin.readline().strip()
      if nextline == "q":
         mtt.Destroy()
         print "Bye bye!"
         sys.exit()
      else: 
         m = Message.Message(666)
         m.PutString("special message!", nextline)
         mtt.SendOutgoingMessage(m)

