#include "python/PythonUtilityFunctions.h"
#include "message/Message.h"

namespace muscle {

static const char * fname(const String & key, const char * defaultFieldName) {return (key.Length() > 0) ? key() : defaultFieldName;}

static status_t ParsePythonDictionary(PyObject * keywords, Message & msg);  // forward declaration
static status_t ParsePythonSequence(  PyObject * args,     Message & msg);  // forward declaration

const char * GetDefaultPythonArgFieldName(type_code type)
{
   switch(type)
   {
      case B_BOOL_TYPE:   case B_INT8_TYPE:   case B_INT16_TYPE: case B_INT32_TYPE: return "_argInt32";
      case B_FLOAT_TYPE:  case B_DOUBLE_TYPE:                                       return "_argFloat";
      case B_STRING_TYPE:                                                           return "_argString";
      case B_POINT_TYPE:                                                            return "_argPoint";
      case B_MESSAGE_TYPE:                                                          return "_argMessage";
      default:                                                                      return NULL;
   }
}

static status_t PythonAddArg(const String & key, PyObject * value, Message & msg)
{
   status_t ret = B_ERROR;

   if (value == Py_None) return B_NO_ERROR;  // None is the same as not specifying the arg

        if (PyInt_Check    (value)) ret = msg.AddInt32( fname(key, "_argInt32"),  PyInt_AS_LONG(value));
   else if (PyLong_Check   (value)) ret = msg.AddInt64( fname(key, "_argInt64"),  PyLong_AsLongLong(value));
   else if (PyFloat_Check  (value)) ret = msg.AddFloat( fname(key, "_argFloat"),  (float)PyFloat_AsDouble(value));
   else if (PyString_Check (value)) ret = msg.AddString(fname(key, "_argString"), PyString_AsString(value));
   else if (PyComplex_Check(value)) ret = msg.AddPoint( fname(key, "_argPoint"),  Point((float)PyComplex_RealAsDouble(value), (float)PyComplex_ImagAsDouble(value)));
   else if (PyUnicode_Check(value)) 
   {
      PyObject * utf8 = PyUnicode_AsUTF8String(value);
      if (utf8) 
      {
         ret = msg.AddString(fname(key, "_argString"), PyString_AsString(utf8));
         Py_DECREF(utf8);
      }
      else ret = B_ERROR;
   }
   else if (PyDict_Check(value))
   {
      MessageRef subMsg = GetMessageFromPool();
      if ((subMsg())&&(msg.AddMessage(fname(key, "_argMessage"), subMsg) == B_NO_ERROR)) ret = ParsePythonDictionary(value, *subMsg());
   }
   else if (PySequence_Check(value))
   {
      MessageRef subMsg = GetMessageFromPool();
      if ((subMsg())&&(msg.AddMessage(fname(key, "_argMessage"), subMsg) == B_NO_ERROR)) ret = ParsePythonSequence(value, *subMsg());
   }
   return ret;
}

static status_t ParsePythonSequence(PyObject * args, Message & msg)
{
   int seqLen = PySequence_Fast_GET_SIZE(args);
   for (int i=0; i<seqLen; i++) if (PythonAddArg("", PySequence_Fast_GET_ITEM(args, i), msg) != B_NO_ERROR) return B_ERROR;
   return B_NO_ERROR;
}

static status_t ParsePythonDictionary(PyObject * keywords, Message & msg)
{
   status_t ret = B_NO_ERROR;
   PyListObject * keys = (PyListObject *) PyDict_Keys(keywords);
   if (keys)
   {
      PyListObject * values = (PyListObject *) PyDict_Values(keywords);
      if (values)
      {
         int numKeys   = PyList_GET_SIZE(keys);
         int numValues = PyList_GET_SIZE(values);
         if (numKeys == numValues)
         {
            ret = B_NO_ERROR;
            for (int i=0; i<numKeys; i++)
            {
               PyObject * key = PyList_GET_ITEM(keys, i);
               if (PyString_Check(key))
               {
                  PyObject * value = PyList_GET_ITEM(values, i);
                  if (PythonAddArg(PyString_AS_STRING(key), value, msg) != B_NO_ERROR)
                  {
                     ret = B_ERROR;
                     break;
                  }
               }
            }
         }
         Py_DECREF(values);
      }
      Py_DECREF(keys);
   }
   return ret;
}


status_t ParsePythonArgs(PyObject * args, PyObject * keywords, Message & msg)
{
   status_t ret = B_NO_ERROR;
   if ((ret == B_NO_ERROR)&&(args)    &&(PySequence_Check(args))) ret = ParsePythonSequence(args, msg);
   if ((ret == B_NO_ERROR)&&(keywords)&&(PyDict_Check(keywords))) ret = ParsePythonDictionary(keywords, msg);
   if (ret != B_NO_ERROR) PyErr_SetString(PyExc_RuntimeError, "Error parsing args into Message format");
   return ret;
}

PyObject * ConvertMessageItemToPyObject(const Message & msg, const String & fieldName, uint32 index)
{
   type_code type;
   if (msg.GetInfo(fieldName, &type) == B_NO_ERROR)
   {
      bool setErr = false;
      switch(type)
      {
         case B_BOOL_TYPE: 
         {
            bool temp;
            if (msg.FindBool(fieldName, index, &temp) == B_NO_ERROR) return PyInt_FromLong(temp?1:0);
         }
         break;

         case B_DOUBLE_TYPE: 
         {
            double temp;
            if (msg.FindDouble(fieldName, index, &temp) == B_NO_ERROR) return PyFloat_FromDouble(temp);
         }
         break;

         case B_FLOAT_TYPE: 
         {
            float temp;
            if (msg.FindFloat(fieldName, index, &temp) == B_NO_ERROR) return PyFloat_FromDouble(temp);
         }
         break;

         case B_INT64_TYPE: 
         {
            int64 temp;
            if (msg.FindInt64(fieldName, index, &temp) == B_NO_ERROR) return PyLong_FromLongLong(temp);
         }
         break;

         case B_INT32_TYPE: 
         {
            int32 temp;
            if (msg.FindInt32(fieldName, index, &temp) == B_NO_ERROR) return PyInt_FromLong(temp);
         }
         break;

         case B_INT16_TYPE: 
         {
            int16 temp;
            if (msg.FindInt16(fieldName, index, &temp) == B_NO_ERROR) return PyInt_FromLong(temp);
         }
         break;

         case B_INT8_TYPE: 
         {
            int8 temp;
            if (msg.FindInt8(fieldName, index, &temp) == B_NO_ERROR) return PyInt_FromLong(temp);
         }
         break;

         case B_POINT_TYPE: 
         {
            Point temp;
            if (msg.FindPoint(fieldName, index, temp) == B_NO_ERROR) return PyComplex_FromDoubles(temp[0], temp[1]);
         }
         break;

         case B_STRING_TYPE: 
         {
            const char * temp;
            if (msg.FindString(fieldName, index, &temp) == B_NO_ERROR) return PyString_FromString(temp);
         }
         break;

         case B_RAW_TYPE: 
         {
            const char * temp;
            uint32 size;
            if (msg.FindData(fieldName, index, B_RAW_TYPE, (const void **)&temp, &size) == B_NO_ERROR) return PyString_FromStringAndSize(temp, (int)size);
         }
         break;

         default:
            PyErr_SetString(PyExc_RuntimeError, "Message contained unsupported datatype");
            setErr = true;
         break;
      }
      if (setErr == false) PyErr_SetString(PyExc_RuntimeError, "Message item not found");
   }
   else PyErr_SetString(PyExc_RuntimeError, "Malformed Message");

   return NULL;
}

}; // end namespace muscle

